//RDA includes
#include <RDAService.h>
#include <Config.h>
#include <DeviceHandle.h>

//misc includdes
#include <iostream>
#include <string>
#include <errno.h>

// The callback object that is used to handle subscription results
class SubscriptionReportHandler: public rdaReplyHandler
{
    public:

        virtual void handleReply(const rdaRequest& rq, const rdaData& value)
        {
            printf("%s/%s = ", rq.getDeviceHandle()->getDeviceName(), rq.getPropertyName());
            value.print();
        }

        virtual void handleError(const rdaRequest& rq, const rdaException& ex)
        {
            printf("%s/%s - error reported\n", rq.getDeviceHandle()->getDeviceName(), rq.getPropertyName());
            std::cout << ex.getType() << "  " << ex.getMessage() << std::endl;
        }

        virtual void disconnected(const rdaRequest& rq)
        {
            printf("%s/%s - disconnected from data source\n", rq.getDeviceHandle()->getDeviceName(),
                            rq.getPropertyName());
        }

        virtual void reconnected(const rdaRequest& rq)
        {
            printf("%s/%s - reconnected to data source\n", rq.getDeviceHandle()->getDeviceName(), rq.getPropertyName());
        }

        virtual void cancelled(const rdaRequest& rq)
        {
            printf("%s/%s - subscription cancelled\n", rq.getDeviceHandle()->getDeviceName(), rq.getPropertyName());
        }
};

// Executes the get operation on the specified property and prints results
void doGet(rdaDeviceHandle* device, const char* prop, const char* cycle)
{
    try
    {
        rdaData context;
        //double sd = 5;
        //context->insert("AlexFilter",sd);
        rdaData* value = device->get(prop, cycle, context);
        value->print();
        delete value;
        std::cout << " action finished without errors" << std::endl;
    }
    catch (const rdaException& ex)
    {
        std::cout << ex.getType() << "  " << ex.getMessage() << std::endl;
    }
    catch (...)
    {
        std::cout << "Unknown exception caught !!! " << std::endl;
    }
}

// Attempts to set the property value
void doSet(rdaDeviceHandle* device, const char* prop, const char* cycle)
{
    try
    {
        rdaData context;
        rdaData data;
        double value = 4711;
        std::cout << "Trying to set the field to: " << value << std::endl;
        data.insert("AnyField",value);
        device->set(prop, cycle, data, context);
        std::cout << " action finished without errors" << std::endl;
    }
    catch (const rdaException& ex)
    {
        std::cout << ex.getType() << "  " << ex.getMessage() << std::endl;
    }
}

// Starts a subscription to the specified property
rdaRequest* subscribe(rdaDeviceHandle* device, const char* prop, const char* cycle,SubscriptionReportHandler& replyHandler)
{
    try
    {
        rdaData* context;
        context = new rdaData();
	rdaRequest* request = NULL;
        //context->insert("testitem","AlexTestValue");
        request = device->monitorOn(prop, cycle, false, &replyHandler, *context);
        //the true is for onchange  = on
        std::cout << "Subscribe finished without errors" << std::endl;
	return request;
    }
    catch (const rdaException& ex)
    {
        std::cout << ex.getType() << "  " << ex.getMessage() << std::endl;
	return NULL;
    }
	 
}

void unsubscribe(rdaDeviceHandle* device, const char* prop, const char* cycle,rdaRequest* request)
{
    try
    {
        rdaData* context;
        context = new rdaData();
        device->monitorOff(request);
        std::cout << "Unsubsribe finished without errors" << std::endl;
    }
    catch (const rdaException& ex)
    {
        std::cout << ex.getType() << "  " << ex.getMessage() << std::endl;
    }
}

int main(int argc, char **argv)
{
    //format is EquipmentName.FECName
    //If the client is outside the acc network, use FEC.acc.gsi.de as FECName

    std::string ServerName = "LocalTestDeployUnit.asl724";

    std::string property_name_GetSet = "NameOfMyProperty";
    std::string property_name_subscription = "NameOfMyProperty";
    std::string device_name = "NameOfMyDevice";
    std::string cycle_selector_subscription = "";
    std::string cycle_name_GetSet = "";

    try
    {
    	std::cout << "---------------------------------------------------------------"<< std::endl;
    	std::cout << "Before using this client, please make sure that you executed the FESA3 environment script:"<<std::endl;
	std::cout << "asl72x:/acc/local/share/fesa/[version]/fesa3_environment.conf"<< std::endl;
        std::cout << "---------------------------------------------------------------"<< std::endl;
        std::cout << "Device: " << device_name << std::endl;
        std::cout << "Property  for Get/Set: " << property_name_GetSet << std::endl;
        std::cout << "CycleName for Get/Set: " << cycle_name_GetSet << std::endl;
        std::cout << "Property for Subscription: " << property_name_subscription << std::endl;
        std::cout << "CycleSelector-Subscription: " << cycle_selector_subscription << std::endl;
	std::cout << "---------------------------------------------------------------"<< std::endl;

    	// Forward all device calls to the specified server
    	rdaConfig::setProperty("cmw.rda.testServer", ServerName.c_str());
		std::cout << "Connected to: " << ServerName << std::endl;

    	// handle to the current subscription request (generated by Server on calling "Subscribe")
    	rdaRequest* request = NULL;

    	//a reply Handler to handle the reply of our subscription
    	SubscriptionReportHandler replyHandler;

    	// handle to our device
    	rdaDeviceHandle* device = 0;

    	rdaRDAService* rda = rdaRDAService::init();

		try
		{
			device = rda->getDeviceHandle(device_name.c_str(),ServerName.c_str());
		}
		catch (...)
		{
			printf("Device not found\n");
			throw;
		}

        char input = ' ';
        while (input != '0')
        { //a very basic Menu

            std::cout << "---------------------------------------------------------------"<< std::endl;
            std::cout << "choose your action:" << std::endl;
            std::cout << "(1) get" << std::endl;
            std::cout << "(2) set" << std::endl;
            std::cout << "(3) add Subscriber" << std::endl;
            std::cout << "(4) remove Subscriber" << std::endl;
            std::cout << "(0) exit" << std::endl;
            std::cout << "---------------------------------------------------------------"<< std::endl;
	    std::cout << "# ";
            std::cin >> input;
            std::cout << "---------------------------------------------------------------"<< std::endl;

            switch (input)
            {
            case '1':
                doGet(device,property_name_GetSet.c_str(), cycle_name_GetSet.c_str());
                break;
            case '2':
                doSet(device,property_name_GetSet.c_str(), cycle_name_GetSet.c_str());
                break;
            case '3':
		if(request != NULL)
		{	
			std::cout << "Cannot handle more than one Subscriber. Please first unsubscribe." << std::endl;
			break;
		}
                request = subscribe(device,property_name_subscription.c_str(), cycle_selector_subscription.c_str(),replyHandler);
                break;
            case '4':
		if(request == NULL)
		{
			std::cout << "There is no active subscription" << std::endl;
		}
		else
		{
			unsubscribe(device,property_name_subscription.c_str(), cycle_selector_subscription.c_str(),request);
			request = NULL;
		}
			
                	
                break;
            case '0':
                break;
            default:
                std::cout << "Invalid Command" << std::endl;
            }

        } // end while
    }
    catch (const rdaInternalException& ex)
    {
        printf("RDA internal exception: %s\n", ex.getMessage());
        return 1;
    }
    catch (const rdaInternalError& ex)
    {
        printf("RDA internal error [%s:%d]: %s\n", ex.getFile(), ex.getLine(), ex.getMessage());
        return 1;
    }
    catch (...)
    {
        printf("Cannot proceed: unknown exception\n");
        return 1;
    }
    return 0;
}

